#include <cryptoauthlib.h>
#include "basic/atca_basic.h"

#define BOOL_TO_ONOFF(b) (b ? "on" : "off")
#define NIBBLE_TO_HEX_CHAR(i) ((i <= 9) ? ('0' + i) : ('A' - 10 + i))
#define HIGH_NIBBLE(i) ((i >> 4) & 0x0F)
#define LOW_NIBBLE(i) (i & 0x0F)

#define HEX_CHAR_TO_NIBBLE(c) ((c >= 'A') ? (c - 'A' + 0x0A) : (c - '0'))
#define HEX_PAIR_TO_BYTE(h, l) ((HEX_CHAR_TO_NIBBLE(h) << 4) + HEX_CHAR_TO_NIBBLE(l))

#define IS_HEX_CHAR(c) ( (((c >= 'A') && (c <= 'F')) || ((c >= '0') && (c <= '9'))) ? true : false)
#define ARRAY_SIZE(a) (sizeof(a)/sizeof(a[0]))

#define debugSerial SerialUSB
#define debugBaud 57600
#define debugDelay 10000

// Key used for encrypted reads and writes 
// between the ATECC chip and the MCU
uint8_t SLOT_4_KEY[ATCA_BLOCK_SIZE] = {
  0x04, 0xE9, 0x97, 0x98, 0x46, 0x5F, 0xB4, 0xA2, 
  0xEC, 0x4E, 0x38, 0xE0, 0xE1, 0x58, 0x54, 0x14, 
  0x33, 0x83, 0x94, 0x39, 0x8F, 0xD8, 0xA7, 0x99, 
  0x1E, 0xA8, 0xD4, 0x63, 0x1B, 0xC9, 0x39, 0xCA
};

// Key fake, with only the first byte altered
// when compared to the original key
uint8_t SLOT_4_KEY_FAKE[ATCA_BLOCK_SIZE] = {
  0x05, 0xE9, 0x97, 0x98, 0x46, 0x5F, 0xB4, 0xA2, 
  0xEC, 0x4E, 0x38, 0xE0, 0xE1, 0x58, 0x54, 0x14, 
  0x33, 0x83, 0x94, 0x39, 0x8F, 0xD8, 0xA7, 0x99, 
  0x1E, 0xA8, 0xD4, 0x63, 0x1B, 0xC9, 0x39, 0xCA
};

uint8_t SOME_SECRET_DATA[ATCA_BLOCK_SIZE] = {
  0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
  0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
  0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
  0x18, 0x19, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F
};

void setup() 
{
  while ((!debugSerial) && (millis() < debugDelay)){
    // Wait debugDelay milliseconds for debugSerial to open
  }

  // Start debug stream
  debugSerial.begin(debugBaud);
  debugSerial.println("Starting...");
  debugSerial.println();
  
  //Init chip
  cfg_ateccx08a_i2c_default.iface.atcai2c.bus = 0;
  debugSerial.print("Initialising ATECCX08A:...");
  showResult(atcab_init(&cfg_ateccx08a_i2c_default));
  debugSerial.println();

  //Read configuration lock status
  bool isLocked = false;
  debugSerial.print("Reading configuration lock status:...");
  showResult(atcab_is_locked(LOCK_ZONE_CONFIG, &isLocked));
  debugSerial.print("The configuration zone is:...");
  debugSerial.println((isLocked ? "LOCKED" : "UNLOCKED"));
  debugSerial.println();

  if (isLocked) {
    // For clarity
    uint8_t SLOT_WITH_KEY = 4;
    uint8_t SLOT_WITHOUT_KEY = 0;
    uint8_t SLOT_WITH_DATA = 8;
    
    // Write encryption key to Slot 4
    debugSerial.print("Writing ecryption key to Slot 4:...");
    showResult(atcab_write_zone(ATCA_ZONE_DATA, SLOT_WITH_KEY, 0, 0, SLOT_4_KEY, ATCA_BLOCK_SIZE));
    debugSerial.println();

    // Note: In production, Slot 4 should now be locked.
    // This would prevent any modification to the encyrption key
    // used for Slot 8. However, the lock is a one time only process.

    // Write some data using the encrypted write method to Slot 8
    debugSerial.print("Writing encrypted DATA to Slot 8:...");
    showResult(atcab_write_enc(SLOT_WITH_DATA, 0, SOME_SECRET_DATA, SLOT_4_KEY, SLOT_WITH_KEY));
    printHex(SOME_SECRET_DATA, sizeof(SOME_SECRET_DATA), 8);
    debugSerial.println();

    // Prepare a buffer for the data read back and initialise to all 0s
    uint8_t READ_DATA[ATCA_BLOCK_SIZE];
    uint8_t READ_FAKE[ATCA_BLOCK_SIZE];
    memset(READ_DATA, 0, sizeof(READ_DATA));
    memset(READ_FAKE, 0, sizeof(READ_FAKE));

    // Read back the data using the encrypted read method
    debugSerial.println("Reading with correct key:");
    debugSerial.print("Reading encrypted DATA from Slot 8:...");
    showResult(atcab_read_enc(SLOT_WITH_DATA, 0, READ_DATA, SLOT_4_KEY, SLOT_WITH_KEY));
    printHex(READ_DATA, sizeof(READ_DATA), 8);
    debugSerial.println();
    
    // Read back the data using the encrypted read method with incorrect key
    // As expected, a small change in key returns unusable data
    debugSerial.println("Reading with incorrect key:");
    debugSerial.print("Reading encrypted DATA from Slot 8:...");
    showResult(atcab_read_enc(SLOT_WITH_DATA, 0, READ_FAKE, SLOT_4_KEY_FAKE, SLOT_WITH_KEY));
    printHex(READ_FAKE, sizeof(READ_FAKE), 8);
    debugSerial.println();
  }
  else {
    debugSerial.println("Device is must first be configured...");
    debugSerial.println();
  }
}

void loop() 
{ 
}

void printHex(const uint8_t* buff, uint8_t len, uint8_t width)
{
    char strA[16];
    for (uint8_t i = 0; i < len; i++)
    {
        if (i % width == 0)
        {
          sprintf(strA,"0x%04x",i);
          debugSerial.print(String((i == 0 ? "" : "\r\n")) + String(strA) + ": ");
        }
        sprintf(strA,"0x%02X, ",buff[i]);
        debugSerial.print(strA);
        
    }
    debugSerial.println();
}

void printRawHex(const uint8_t* buff, uint8_t len)
{
    for (uint8_t i = 0; i < len; i++)
    {
      debugSerial.print((char)NIBBLE_TO_HEX_CHAR(HIGH_NIBBLE(buff[i])));
      debugSerial.print((char)NIBBLE_TO_HEX_CHAR(LOW_NIBBLE(buff[i])));
    }
    debugSerial.println();
}

void halt()
{
  while (true) {
    delay(1000);
  }
}

void showResult(ATCA_STATUS result)
{
  String message;
  
  switch(result) {
    case ATCA_SUCCESS                 :message = "SUCCESS"; break;
    case ATCA_CONFIG_ZONE_LOCKED      :message = "CONFIG_ZONE_LOCKED"; break;
    case ATCA_DATA_ZONE_LOCKED        :message = "DATA_ZONE_LOCKED"; break;
    case ATCA_WAKE_FAILED             :message = "WAKE_FAILED"; break;
    case ATCA_CHECKMAC_VERIFY_FAILED  :message = "CHECKMAC_VERIFY_FAILED"; break;
    case ATCA_PARSE_ERROR             :message = "PARSE_ERROR"; break;
    case ATCA_STATUS_CRC              :message = "STATUS_CRC"; break;
    case ATCA_STATUS_UNKNOWN          :message = "STATUS_UNKNOWN"; break;
    case ATCA_STATUS_ECC              :message = "STATUS_ECC"; break;
    case ATCA_FUNC_FAIL               :message = "FUNC_FAIL"; break;
    case ATCA_GEN_FAIL                :message = "GEN_FAIL"; break;
    case ATCA_BAD_PARAM               :message = "BAD_PARAM"; break;
    case ATCA_INVALID_ID              :message = "INVALID_ID"; break;
    case ATCA_INVALID_SIZE            :message = "INVALID_SIZE"; break;
    case ATCA_RX_FAIL                 :message = "RX_FAIL"; break;
    case ATCA_RX_NO_RESPONSE          :message = "RX_NO_RESPONSE"; break;
    case ATCA_RESYNC_WITH_WAKEUP      :message = "RESYNC_WITH_WAKEUP"; break;
    case ATCA_PARITY_ERROR            :message = "PARITY_ERROR"; break;
    case ATCA_TX_TIMEOUT              :message = "TX_TIMEOUT"; break;
    case ATCA_RX_TIMEOUT              :message = "RX_TIMEOUT"; break;
    case ATCA_COMM_FAIL               :message = "COMM_FAIL"; break;
    case ATCA_TIMEOUT                 :message = "TIMEOUT"; break;
    case ATCA_BAD_OPCODE              :message = "BAD_OPCODE"; break;
    case ATCA_WAKE_SUCCESS            :message = "WAKE_SUCCESS"; break;
    case ATCA_EXECUTION_ERROR         :message = "EXECUTION_ERROR"; break;
    case ATCA_UNIMPLEMENTED           :message = "UNIMPLEMENTED"; break;
    case ATCA_ASSERT_FAILURE          :message = "ASSERT_FAILURE"; break;
    case ATCA_TX_FAIL                 :message = "TX_FAIL"; break;
    case ATCA_NOT_LOCKED              :message = "NOT_LOCKED"; break;
    case ATCA_NO_DEVICES              :message = "NO_DEVICES"; break;
  }

  debugSerial.println(message);

  if (result != ATCA_SUCCESS) {
    halt();
  }
}
